Require Import Peano_dec.
Require Import Compare_dec.
Require Import Lt.
Require Import Le.
Require Import Gt.
Require Import Plus.
Require Import Minus.
Require Import Bool.
Require Import base.
Require Import term.


Inductive Beta : Term -> Term -> Prop :=
 | Beta_head : forall  A b B c,  Beta ((λ[ A ], b)·(B) c)  ( b [← c])
 | Beta_red1 : forall x B y z, Beta x  y ->Beta (x·(B) z)  (y·(B) z)
 | Beta_red2 : forall x B y z, Beta x  y -> Beta (z·(B) x)  (z ·(B) y)
 | Beta_red3 : forall x y B B', Beta B B' -> Beta (x·(B) y)  (x ·(B') y)
 | Beta_lam : forall x y A , Beta x  y -> Beta (λ [A],  x)  (λ [ A], y)
 | Beta_lam2 : forall x A B , Beta A  B -> Beta (λ [ A ], x)  (λ [B ], x)
 | Beta_pi : forall x y A ,Beta  x  y -> Beta (Π (A), x)  (Π(A), y)
 | Beta_pi2 : forall x A B, Beta A  B -> Beta (Π(A), x)  (Π(B), x)
.

Inductive Betas : Term -> Term -> Prop :=
 | Betas_refl : forall x, Betas x x
 | Betas_Beta : forall x y, Beta x y -> Betas x y
 | Betas_trans : forall x y z, Betas x y -> Betas y z -> Betas x z.

Hint Constructors Beta.
Hint Constructors Betas.


Notation " A → B " := (Beta A B) (at level 80) : Typ_scope.
Notation " A →→ B " := (Betas A B) (at level 80) : Typ_scope.


Lemma Betas_App : forall a b B B' c d, a →→ b -> B →→ B' -> c →→ d -> 
  a·(B)c →→ b·(B')d.
assert (forall a b c B , b →→ c ->  a·(B)b →→ a·(B)c).
induction 1; eauto.
assert (forall a b c B, b→→c -> b·(B) a →→ c·(B) a).
induction 1; eauto.
assert (forall a b B B', B →→ B' -> a·(B) b →→ a·(B') b).
induction 1; eauto.

intros. eapply Betas_trans. apply H. apply H4. eapply Betas_trans. apply H0.
apply H2. apply H1. trivial.
Qed.

Hint Resolve Betas_App.

Lemma Betas_La : forall a b c d, a →→  b -> c →→ d -> λ[a], c →→ λ[b], d.
assert (forall a b c , a →→ b -> λ[c],  a →→ λ[c], b).
induction 1; eauto.
assert (forall a b c, a →→ b -> λ[a], c →→ λ[b], c).
induction 1; eauto.
eauto.
Qed.


Hint Resolve Betas_La.


Lemma Betas_Pi : forall a b c d, a →→ b -> c →→ d -> Π(a), c →→ Π(b), d.
assert (forall a b c , a →→ b -> Π (c), a →→ Π(c), b).
induction 1; eauto.
assert (forall a b c, a →→ b -> Π(a), c →→ Π(b), c).
induction 1; eauto.
eauto.
Qed.

Hint Resolve Betas_Pi.


Inductive Betap : Term -> Term -> Prop :=
 | Betap_sort : forall s, Betap !s !s
 | Betap_var : forall x, Betap #x #x
 | Betap_lam: forall m m' A A', Betap m m' ->Betap A A'-> 
    Betap (λ [ A], m) (λ[ A'], m')
 | Betap_pi: forall m m' A A', Betap m m' -> Betap A A' -> 
    Betap (Π(A), m) (Π(A'), m')
 | Betap_App : forall m m' U U' n n', Betap m m' -> Betap n n' -> 
 Betap U U' -> Betap (m ·(U) n) (m'·(U') n')
 | Betap_head : forall m m' n n' A B , Betap m m' -> Betap n n' -> 
    Betap ((λ[A], m)·(B) n) (m'[← n']).

Notation "m →' n" := (Betap m n) (at level 80) : Typ_scope.

Hint Constructors Betap.

Lemma Betap_lift: forall a b n m, a →' b -> a ↑ n # m →' b ↑ n # m .
intros.
revert n m.
induction H; simpl; intuition.
destruct le_gt_dec; intuition.
change m0 with (0+m0).
rewrite substP1.
constructor; intuition.
Qed.

Hint Resolve Betap_lift.

(* inversion lemmas for beta // *)
Lemma Betap1:forall m m' n n' x, 
  m →' m'  -> n →' n' -> m[x←n] →' m'[x←n'].
intros.
revert x.
induction H; simpl; intuition.
destruct lt_eq_lt_dec as [[] | ]; intuition.
rewrite subst_travers.  replace (S x) with (x+1). constructor; intuition.
rewrite <- plus_Snm_nSm. rewrite plus_n_O. trivial.
Qed.

(* Beta // has the diamond property *)
Lemma Betap_diamond : forall m m1 m2 ,
   m →' m1 -> m →' m2 -> (exists m3, m1 →' m3 /\ m2 →' m3). 
intros.
revert m2 H0. 
induction H; intros.
(*1/5*)
inversion H0; subst; clear H0.
exists !s; intuition.
(**)
inversion H0; subst; clear H0.
exists #x; intuition.
(*2/5*)
inversion H1; subst; clear H1. 
destruct (IHBetap1 m'0 H4) as (b & ? & ?).
destruct (IHBetap2 A'0 H6) as (D & ?& ?).
exists (λ[D], b); intuition. 
(*3/5*)
inversion H1; subst; clear H1. 
destruct (IHBetap1 m'0 H4) as (B'' & ? & ?).
destruct (IHBetap2 A'0 H6) as (A'' & ?& ?).
exists (Π(A''), B''); intuition. 
(*4/5*)
inversion H2; subst; clear H2.
destruct (IHBetap1 m'0 H6) as (M & ? & ?); clear IHBetap1.
destruct (IHBetap2 n'0 H8) as (N & ? & ?); clear IHBetap2.
destruct (IHBetap3 U'0 H9) as (B & ? & ?); clear IHBetap3.
exists (App M B N); intuition.
destruct (IHBetap2 n'0 H8) as (N & ? & ?); clear IHBetap2.
clear IHBetap3.
inversion H; subst; clear H.
destruct (IHBetap1 (La A' m'0)) as (L & ? & ?); clear IHBetap1.
constructor; intuition. inversion H; subst; clear H.
inversion H4; subst; clear H4. exists (m' [ ← N]); intuition.
apply Betap1. trivial. trivial.
(**)
inversion H1; subst; clear H1. inversion H5; subst; clear H5.
destruct (IHBetap1 m'1 H3) as (M & ? & ?).
destruct (IHBetap2 n'0 H7) as (N & ? & ?).
exists (M[ ← N]); intuition. apply Betap1; trivial.
destruct (IHBetap1 m'0 H7) as (M & ? & ?).
destruct (IHBetap2 n'0 H8) as (N & ? & ?).
exists (M[ ← N]); intuition. apply Betap1; trivial. apply Betap1; trivial.
Qed.



Lemma Betap_to_Betas : forall M N, Betap M N -> Betas M N.
induction 1; intuition.
eapply Betas_trans. eapply Betas_App. eapply Betas_La. apply Betas_refl.
apply IHBetap1. apply Betas_refl. apply IHBetap2. constructor.
constructor.
Qed.




Inductive Betaps : Term -> Term -> Prop :=
 | Betaps_refl : forall x , Betaps  x x
 | Betaps_trans : forall x y z  , Betap x y -> Betaps y z -> Betaps x z
.


Notation " x  →→' y " := (Betaps x y) (at level 80) : Typ_scope.

Hint Constructors Betaps.


Lemma Betas_Betap_closure : forall x y , x →' y -> x →→ y.
induction 1; eauto.
Qed.

Hint Resolve Betas_Betap_closure.


Lemma Betas_Betaps_closure : forall x y , 
  x →→'  y -> x →→  y.
induction 1; eauto.
Qed.

Lemma Betap_refl : forall x, Betap x x.
induction x; intuition.
Qed.

Hint Resolve Betap_refl.


Lemma Betap_Beta_closure : forall x y,
 x → y -> x →' y.
induction 1; intuition.
Qed.

Local Hint Resolve Betas_Betaps_closure Betap_Beta_closure.

Lemma Betaps_Beta_closure :forall x y, x → y -> x →→' y.
eauto.
Qed.

Hint Resolve Betaps_Beta_closure.

Lemma Betaps_trans2 : forall x y z , x →→' y -> y →→' z  -> x →→' z.
intros. revert z H0; induction H; eauto.
Qed.

Hint Resolve Betaps_trans2.

Lemma Betaps_Betas_closure : forall x y , x →→ y -> x →→' y.
induction 1; eauto.
Qed.


Lemma Betap_diamond_ : forall m m1 m2 ,
   m →→' m1 -> m →' m2 -> (exists m3, m1 →' m3 /\ m2 →→' m3).
intros until 1. revert m2. induction H; intros.
exists m2; intuition.
destruct (Betap_diamond x y m2 H H1) as (q & ? &?).
destruct (IHBetaps q H2) as (q' & ? & ?). exists q'; eauto.
Qed.

Lemma Betaps_diamond : forall m m1 m2 ,
   m →→' m1 -> m →→' m2 -> (exists m3, m1 →→' m3 /\ m2 →→' m3).
intros. revert m1 H. induction H0; intros.
exists m1; intuition.
destruct (Betap_diamond_ x m1 y H1 H) as (q & ? & ? ).
destruct (IHBetaps q H3) as (q' & ?& ?).
exists q'; eauto.
Qed.

Lemma Betas_diamond : forall x y, x →→ y -> forall z, x →→ z -> exists q, y →→ q /\ z →→ q.
intros.
apply Betaps_Betas_closure in H. apply Betaps_Betas_closure in H0.
destruct (Betaps_diamond x y z H H0) as (q & ? & ?).
exists q; split; apply Betas_Betaps_closure; trivial.
Qed.


