Require Import base.
Require Import ut_term.
Require Import ut_red.
Require Import term.
Require Import env.
Require Import Peano_dec.
Require Import Compare_dec.
Require Import Lt.
Require Import List.


Open Scope Typ_scope.

Fixpoint strip (T : term.Term ) : ut_term.Term := match T with
 | Pi A B => ut_term.Pi (strip A) (strip B)
 | La A M => ut_term.La (strip A) (strip M)
 | App M B N => ut_term.App (strip M) (strip N)
 | !s => ut_term.Sort s
 | #x => ut_term.Var x
end.

Lemma strip_lift : forall M n m, strip (M ↑ n # m) = ((strip M) ↑ n # m)%UT.
induction M; intros; simpl in *.
destruct le_gt_dec; simpl; trivial.
trivial.
rewrite IHM1, IHM3; trivial.
rewrite IHM1, IHM2; trivial.
rewrite IHM1, IHM2; trivial.
Qed.

Lemma strip_subst : forall M n N, strip (M [n ← N]) = ((strip M) [ n ← strip N])%UT.
induction M; intros; simpl in *.
destruct lt_eq_lt_dec as [ [] | ]; simpl; trivial.
rewrite strip_lift. trivial.
trivial.
rewrite IHM1, IHM3; trivial.
rewrite IHM1, IHM2; trivial.
rewrite IHM1, IHM2; trivial.
Qed.

Fixpoint strip_env L := match L with
 | nil => nil
 | cons t q => cons (strip t) (strip_env q)
end.


Lemma strip_Tels : forall l t, strip (Tels l t) = (ut_red.Tels (strip_env l) (strip t))%UT.
induction l; simpl; intuition.
rewrite IHl. trivial.
Qed.

Lemma strip_Tels' : forall l t, strip (Tels' l t) = (ut_red.Tels' (strip_env l) (strip t))%UT.
induction l; simpl; intuition.
rewrite IHl. trivial.
Qed.
