Require Import Peano_dec.
Require Import Compare_dec.
Require Import Lt.
Require Import Le.
Require Import Gt.
Require Import Plus.
Require Import Minus.
Require Import Bool.
Require Import base.



Inductive Term : Set:=
 | Var : Vars -> Term
 | Sort : Sorts -> Term
 | App : Term -> Term -> Term -> Term
 | Pi : Term -> Term -> Term
 | La :Term -> Term -> Term
.

Notation "x ·( B ) y" := (App x B y) (at level 15, left associativity) : Typ_scope.
Notation "! s" := (Sort s) (at level 1) : Typ_scope.
Notation "# v" := (Var v) (at level 1) : Typ_scope.
Notation "'Π' ( U ) , V " := (Pi U V) (at level 20, U, V at level 30) : Typ_scope.
Notation "'λ' [ U ] , v " := (La U v) (at level 20, U , v at level 30) : Typ_scope.


Reserved Notation " t ↑ x # n " (at level 5, x at level 0, left associativity).

Delimit Scope Typ_scope with Typ. 

Open Scope Typ_scope.

(* lift c n times *)
Fixpoint lift_rec (n:nat) (k:nat) (c':Term) {struct c'} := match c' with
   | # i =>  if le_gt_dec k i then Var (n+i) else Var i
   | ! s => Sort s
   | u ·(B) v=>  App (u ↑ n # k) (B ↑ n # (S k)) (v ↑ n # k)
   | Π ( A ), B => Π (A ↑ n # k), (B ↑ n # (S k))
   | λ [ A ], b => λ [A ↑ n # k], (b ↑ n # (S k)) 
 end  
   where "t ↑ n # k" := (lift_rec n k t) : Typ_scope.

Notation " t ↑ n " := (lift_rec n 0 t) (at level 5, n at level 0, left associativity) : Typ_scope.


Lemma inv_lift : forall A B n m , A ↑ n # m = B ↑ n # m -> A = B.
intros A; induction A; destruct B; intros;
simpl in *; try (discriminate || intuition); (try (destruct (le_gt_dec m v) ; discriminate)).
destruct (le_gt_dec m v); destruct (le_gt_dec m v0) ; injection H; intros; subst; intuition.
apply plus_reg_l in H0; rewrite H0; trivial. 
elim (lt_irrefl m). apply le_lt_trans with v. trivial. induction n; intuition.
elim (lt_irrefl v0). apply lt_le_trans with m. induction n; intuition. trivial.
injection H; intros; rewrite (IHA1 B1 n m H2); rewrite (IHA2 B2 n (S m) H1); rewrite (IHA3 B3 n m H0); reflexivity.
injection H; intros; rewrite (IHA1 B1 n m H1); rewrite (IHA2 B2 n (S m) H0); reflexivity.
injection H; intros; rewrite (IHA1 B1 n m H1); rewrite (IHA2 B2 n (S m) H0); reflexivity.
Qed.

Lemma lift_rec0 : forall c n, c ↑ 0 # n = c.
induction c; intros; simpl.
destruct (le_gt_dec n v); reflexivity.
reflexivity.
rewrite IHc1; rewrite IHc2; rewrite IHc3; reflexivity.
rewrite IHc1; rewrite IHc2; reflexivity.
rewrite IHc1; rewrite IHc2; reflexivity. 
Qed.

Lemma lift0 : forall c, c ↑ 0 = c .
intros; apply lift_rec0.
Qed.

Lemma liftP1 : forall t i j  k, (t ↑ j # i) ↑ k # (j+i) = t ↑ (j+k) # i.
intros t; induction t; intros;simpl.
destruct (le_gt_dec i v); simpl.
destruct (le_gt_dec (j+i) (j+v)); simpl. 
rewrite plus_assoc. replace (k+j) with (j+k) by (apply plus_comm).  trivial. 
apply plus_gt_reg_l in g. elim (lt_irrefl v).
apply lt_le_trans with i; intuition.
simpl; destruct (le_gt_dec (j+i)); intuition.
elim (lt_irrefl v).
apply lt_le_trans with i; intuition. induction j; intuition.
reflexivity.
rewrite IHt1;rewrite <- IHt2;replace (j+S i) with (S(j+i)) by intuition; rewrite IHt3; reflexivity.
rewrite IHt1; rewrite <-IHt2 ;replace (j+S i) with (S(j+i)) by intuition; reflexivity.
rewrite IHt1; rewrite <- IHt2 ;replace (j+S i) with (S(j+i)) by intuition; reflexivity.
Qed.

Lemma liftP2: forall t i j k n, i <= n ->
  (t ↑ j # i) ↑ k # (j+n) = (t ↑ k # n) ↑ j # i.
intro t; induction t; intros; simpl.
destruct (le_gt_dec i v); destruct (le_gt_dec n v).
simpl.
destruct le_gt_dec. destruct le_gt_dec.
rewrite 2! plus_assoc. replace (k+j) with (j+k) by (apply plus_comm).  trivial. 
elim (lt_irrefl v). apply lt_le_trans with i. induction k; intuition. trivial.
apply plus_gt_reg_l in g. elim (lt_irrefl v).
apply lt_le_trans with n; intuition.
simpl.
destruct le_gt_dec. apply plus_le_reg_l in l0. elim (lt_irrefl v).
apply lt_le_trans with n; intuition. destruct le_gt_dec. trivial.
elim (lt_irrefl v). apply lt_le_trans with i; intuition.
simpl. destruct le_gt_dec. elim (lt_irrefl n). apply lt_le_trans with i.
apply le_lt_trans with v; intuition. trivial. elim (lt_irrefl v).
apply lt_le_trans with n. apply lt_le_trans with i; intuition. trivial.
simpl. destruct le_gt_dec. elim (lt_irrefl v). apply lt_le_trans with n.
intuition. induction j; intuition. destruct le_gt_dec. elim (lt_irrefl i).
apply le_lt_trans with v; intuition. trivial.
trivial.

rewrite IHt1; intuition. replace (S(j+n)) with (j+S n) by intuition.
rewrite IHt2; intuition. rewrite IHt3; intuition.
rewrite IHt1; intuition.
replace (S(j+n)) with (j+S n) by intuition.
rewrite (IHt2 (S i) j k (S n)); intuition.
rewrite IHt1; intuition.
replace (S(j+n)) with (j+S n) by intuition.
rewrite (IHt2 (S i) j k (S n) ); intuition.
Qed.

Lemma liftP3 : forall t i k j n , i <= k -> k <= (i+n) ->
  (t ↑ n # i) ↑ j # k = t ↑ (j+n) # i.
intro t; induction t; intros; simpl.
destruct (le_gt_dec i v); simpl.
destruct (le_gt_dec k (n+v)); intuition.
elim (lt_irrefl (i+n)). apply lt_le_trans with k.
apply le_lt_trans with (n+v). rewrite plus_comm. intuition. intuition. trivial.
destruct (le_gt_dec k v); intuition. elim (lt_irrefl k).
apply lt_le_trans with i.  apply le_lt_trans with v. trivial. intuition. trivial.
reflexivity.
rewrite IHt1; intuition. rewrite IHt2; intuition. rewrite IHt3; intuition. change (S i + n) with (S (i+n)). intuition.
rewrite IHt1; intuition;rewrite IHt2; intuition. change (S i + n) with (S (i+n)). intuition.
rewrite IHt1; intuition; rewrite IHt2; intuition. change (S i + n) with (S (i+n)). intuition.
Qed.


Lemma lift_lift : forall t n m, (t ↑ m) ↑ n  = t↑ (n+m).
intros.
apply liftP3; intuition.
Qed.



(* i = index to replace
  t = term we are replacing
 k = limit between free and bound indices
*)
Reserved Notation "t [ x ← u ]" (at level 5, x at level 0, left associativity).

Fixpoint subst_rec w t n {struct t} :=
 match t with
  | # v => match (lt_eq_lt_dec v n) with
      | inleft (left _) => # v (* v < n *)
      | inleft (right _) => w ↑ n  (* v = n *)
      | inright _ => # (v - 1) (* v > n *)
      end
  | ! s => ! s
  | u ·(B) v => (u [ n ← w ]) ·(B [ S n ← w]) ( v [ n ← w ]) 
  | Π ( A ), B => Π ( A [ n ← w ] ), (B [ S n ← w ]) 
  | λ [ A ], b => λ [ A [ n ← w ] ], (b [ S n ← w ]) 
end
    where " t [ n ← w ] " := (subst_rec w t n) : Typ_scope.


Notation " t [ ← w ] " := (subst_rec w t 0) (at level 5) : Typ_scope.


Lemma expand_term_with_subst : forall B n, (B ↑ 1 # (S n)) [ n ← #0 ] = B.
induction B; intros.
unfold lift_rec.
destruct (le_gt_dec (S n) v). unfold subst_rec.
destruct (lt_eq_lt_dec (1+v) n) as [[] | ].
apply le_not_lt in l. elim l. intuition.
elim (lt_irrefl v). apply lt_le_trans with (S (S v)). intuition. subst; trivial.
change (1+v) with (S v). destruct v; simpl; trivial.
simpl.
destruct (lt_eq_lt_dec v n) as [[] | ].
trivial.
simpl; subst; trivial.
rewrite <- plus_n_O. trivial.
elim (lt_irrefl n). apply lt_le_trans with v; intuition.
simpl; trivial.
simpl. rewrite IHB1. rewrite IHB2. rewrite IHB3. reflexivity.
simpl. rewrite IHB1. rewrite IHB2. reflexivity.
simpl. rewrite IHB1. rewrite IHB2. reflexivity.
Qed.


Lemma substP1: forall t u i j k , 
  ( t [ j ← u] ) ↑ k # (j+i) = (t ↑ k # (S (j+i))) [ j ← (u ↑ k # i ) ].
intros t; induction t; intros.
simpl (#v [j ← u] ↑ k # (j+i)).
change (#v ↑ k # (S (j+i))) with  (if le_gt_dec (S (j+i)) v then #(k+v) else #v).
destruct (lt_eq_lt_dec v j) as [[] | ].
destruct (le_gt_dec (S (j+i)) v).
elim (lt_irrefl v). apply lt_le_trans with j; intuition.
apply le_trans with (S (j+i)); intuition.
simpl.
destruct (le_gt_dec (j+i) v).
elim (lt_irrefl v). apply lt_le_trans with j; intuition. apply le_trans with (j+i); intuition.
destruct (lt_eq_lt_dec v j) as [[] | ]. trivial.
subst. elim (lt_irrefl j);trivial.
elim (lt_irrefl j); apply lt_trans with v; trivial.
destruct (le_gt_dec (S(j+i)) v). subst.
elim (lt_irrefl j). apply lt_le_trans with (S (j+i)). intuition. trivial.
simpl. destruct (lt_eq_lt_dec v j) as [[] | ].
subst. elim (lt_irrefl j); trivial.
apply liftP2; intuition.
subst. elim (lt_irrefl j); trivial.
destruct (le_gt_dec (S (j+i)) v).
simpl.
destruct (le_gt_dec (j+i) (v-1)). destruct (lt_eq_lt_dec (k+v) j) as [[] | ].
elim (lt_irrefl j). apply lt_le_trans with v. trivial. induction k; intuition.
subst. elim (lt_irrefl v). apply lt_le_trans with (S(k+v+i)). intuition. trivial.
destruct v. apply lt_n_O in l; elim l. rewrite <- 2! pred_of_minus. replace (k+ S v) with (S (k+v)) by intuition.
simpl. trivial.
elim (lt_irrefl v). apply lt_le_trans with (S (j+i)). destruct v. apply lt_n_O in l; elim l. 
rewrite <- pred_of_minus in g. simpl in g. intuition. trivial.
simpl.
destruct (le_gt_dec (j+i) (v-1)). destruct (lt_eq_lt_dec v j) as [[] | ].
elim (lt_irrefl j); apply lt_trans with v; trivial.
subst. elim (lt_irrefl j); trivial.
elim (lt_irrefl v). apply lt_le_trans with (S (j+i)). intuition. 
destruct v. apply lt_n_O in l; elim l. rewrite <- pred_of_minus in l0. simpl in l0. intuition.
destruct (lt_eq_lt_dec) as [[] | ]. elim (lt_irrefl j); apply lt_trans with v; trivial.
subst. elim (lt_irrefl j); trivial. trivial.
simpl; trivial.
simpl; rewrite IHt1; intuition; rewrite IHt3; intuition.
replace (S(S(j+i))) with (S((S j)+i)) by intuition.
rewrite <- (IHt2 u i (S j)  k ); intuition.
simpl; rewrite IHt1; intuition.
replace (S(S(j+i))) with (S((S j)+i)) by intuition.
rewrite <- (IHt2 u i (S j)  k ); intuition.
simpl; rewrite IHt1; intuition.
replace (S(S(j+i))) with ((S ((S j)+i))) by intuition.
rewrite <- (IHt2 u i (S j)  k ); intuition.
Qed.

Lemma substP2: forall t u i j n, i <= n ->
  (t ↑ j # i ) [ j+n ← u ] = ( t [ n ← u]) ↑ j # i .
intro t; induction t; intros; simpl.
destruct (le_gt_dec i v); destruct (lt_eq_lt_dec v n) as [[] | ].
simpl.
destruct (le_gt_dec i v).  destruct (lt_eq_lt_dec (j+v) (j+n)) as [[] | ].
reflexivity.
apply plus_reg_l in e. subst. elim (lt_irrefl n); trivial.
apply plus_lt_reg_l in l2. elim (lt_asym v n); trivial.
elim (lt_irrefl i). apply le_lt_trans with v; intuition. subst.
simpl.
destruct (lt_eq_lt_dec (j+n) (j+n)) as [[] | ].
apply lt_irrefl in l0; elim l0.
symmetry.
apply liftP3; intuition.
apply lt_irrefl in l0; elim l0.
simpl.
destruct (le_gt_dec i (v-1)). destruct (lt_eq_lt_dec (j+v) (j+n))as [[] | ].
apply plus_lt_reg_l in l2. elim (lt_asym  v n ); trivial.
apply plus_reg_l in e; subst. elim (lt_irrefl n); trivial.
destruct v. apply lt_n_O in l0; elim l0. rewrite <- 2! pred_of_minus. replace (j+ S v) with (S (j+v)) by intuition.
simpl. trivial.
unfold gt in g. unfold lt in g. rewrite <- pred_of_minus in g. 
rewrite <- (S_pred  v n l0) in g.
elim (lt_irrefl n). apply lt_le_trans with v; trivial. apply le_trans with i; trivial.
simpl. 
destruct (le_gt_dec i v).  elim (lt_irrefl i). apply le_lt_trans with v; trivial.
destruct (lt_eq_lt_dec v (j+n)) as [[] | ]. reflexivity.
subst. elim (lt_irrefl n). apply le_lt_trans with (j+n); intuition. 
elim (lt_irrefl n). apply lt_trans with v.  apply le_lt_trans with (j+n); intuition. trivial.
simpl. subst. 
elim (lt_irrefl n). apply lt_le_trans with i; intuition.
simpl. elim (lt_irrefl n). apply lt_le_trans with v; intuition.
apply le_trans with i; intuition.
trivial.

rewrite IHt1; intuition;rewrite IHt3; intuition.
rewrite <- (IHt2 u (S i) j (S n)); intuition.
replace (S(j+n)) with (j+(S n)) by intuition.
reflexivity.
rewrite IHt1; intuition;
rewrite <- (IHt2 u (S i) j (S n)); intuition.
replace (S(j+n)) with (j+(S n)) by intuition.
reflexivity.
rewrite IHt1; intuition;
rewrite <- (IHt2 u (S i) j (S n)); intuition.
replace (S(j+n)) with (j+(S n)) by intuition.
reflexivity.
Qed.


Lemma substP3: forall t u i  k n, i <= k -> k <= i+n ->
  (t↑ (S n) # i) [ k← u] = t ↑ n # i.
intro t; induction t; intros; simpl.
destruct (le_gt_dec i v).
unfold subst_rec.
destruct (lt_eq_lt_dec (S(n+v)) k) as [[] | ].
elim (lt_irrefl (i+n)). apply lt_le_trans with k; intuition.
apply le_lt_trans with (v+n). intuition. rewrite plus_comm; intuition.
subst. replace (i+n) with (n+i) in H0 by (apply plus_comm) . replace (S (n+v)) with (n + S v) in H0 by intuition. 
apply plus_le_reg_l in H0. elim (lt_irrefl i). apply le_lt_trans with v; intuition.
simpl. rewrite <- minus_n_O. trivial.
simpl. destruct (lt_eq_lt_dec v k) as [[] | ].
reflexivity. subst. elim (lt_irrefl i). apply le_lt_trans with k; intuition.
elim (lt_irrefl k). apply lt_trans with v; trivial. apply lt_le_trans with i; intuition.

reflexivity.
rewrite IHt1; intuition;rewrite IHt3; intuition.
rewrite <- (IHt2 u (S i) (S k) n); intuition.
change (S i + n) with (S (i+n)). intuition.
rewrite IHt1; intuition; rewrite <- (IHt2 u (S i) (S k) n); intuition.
change (S i + n) with (S (i+n)). intuition.
rewrite IHt1; intuition; rewrite <- (IHt2 u (S i) (S k) n); intuition.
change (S i + n) with (S (i+n)). intuition.
Qed.

Lemma substP4: forall t u w i j, 
   (t [ i← u]) [i+j ← w] = (t [S(i+j) ← w]) [i← u[j← w]].
intro t; induction t; intros; simpl.
destruct (lt_eq_lt_dec v i) as [[] | ] ; destruct (lt_eq_lt_dec v (S(i+j))) as [[] | ].
simpl.
destruct (lt_eq_lt_dec v (i+j)) as [[] | ]. destruct (lt_eq_lt_dec v i) as [[] | ].
trivial.
subst. apply lt_irrefl in l; elim l. elim ( lt_asym v i); trivial.
subst. rewrite plus_comm in l. elim (lt_irrefl i). induction j; simpl in *; intuition.
elim (lt_irrefl i). apply le_lt_trans with v;intuition. rewrite plus_comm in l1; intuition.  induction j; simpl in *; intuition.
subst. elim (lt_irrefl i). apply lt_trans with (S (i+j)); intuition.
elim (lt_irrefl i). apply lt_trans with (S (i+j)); intuition. apply lt_trans with v; trivial.
simpl. subst. destruct (lt_eq_lt_dec i i) as [[] | ].
elim (lt_irrefl i); trivial. apply substP2; intuition.
elim (lt_irrefl i); trivial.
subst. rewrite plus_comm in e0. apply succ_plus_discr in e0. elim e0.
subst. elim (lt_irrefl i). apply le_lt_trans with (i+j); intuition.
simpl.
destruct (lt_eq_lt_dec (v-1) (i+j)) as  [[] | ]. destruct (lt_eq_lt_dec v i) as [[] | ].
elim (lt_asym v i); trivial. subst. elim (lt_irrefl i); trivial.
trivial. rewrite <- e in l0. rewrite <- pred_of_minus in l0.
rewrite <- (S_pred  v i l) in l0. elim (lt_irrefl v); trivial.
apply lt_n_S in l1. elim (lt_irrefl v).
apply lt_trans with (S(i+j)); trivial.
rewrite <- pred_of_minus in l1. rewrite <- (S_pred  v i l) in l1. trivial. 
subst. simpl. rewrite <- minus_n_O.
destruct (lt_eq_lt_dec (i+j) (i+j)) as [[] | ].
elim (lt_irrefl (i+j)) ; trivial.
symmetry. apply substP3; intuition.
elim (lt_irrefl (i+j)) ; trivial.
simpl.
destruct (lt_eq_lt_dec (v-1) (i+j)) as  [[] | ].
elim (lt_irrefl v). apply lt_trans with (S (i+j)) ;trivial.
apply lt_n_S in l1. rewrite <- pred_of_minus in l1. rewrite <- (S_pred v i l) in l1. trivial.
apply eq_S in e. rewrite <- pred_of_minus in e. rewrite <- (S_pred v i l) in e.
subst. elim (lt_irrefl (S(i+j))); trivial.
destruct (lt_eq_lt_dec (v-1) i) as [[] | ].
elim (lt_irrefl v). apply le_lt_trans with i; trivial. destruct v. apply lt_n_O in l; elim l.
rewrite <- pred_of_minus in l2. simpl in l2. trivial.
destruct v. elim (lt_n_O i); trivial. rewrite <- pred_of_minus in e. simpl in e. subst.
rewrite <- pred_of_minus in l1. simpl in l1. elim (lt_irrefl i).
apply le_lt_trans with (i+j); intuition.
trivial.
trivial.
rewrite IHt1; rewrite IHt3; intuition.
replace (S(S(i+j))) with (S((S i)+ j)) by intuition;
  rewrite <- (IHt2 u w (S i)); replace (S(i+j)) with ((S i)+ j) by intuition; intuition.
rewrite IHt1; replace (S(S(i+j))) with (S((S i)+ j)) by intuition;
  rewrite <- (IHt2 u w (S i)); replace (S(i+j)) with ((S i)+ j) by intuition; intuition.
rewrite IHt1; replace (S(S(i+j))) with (S((S i)+j)) by intuition;
  rewrite <- (IHt2 u w (S i)); replace (S(i+j)) with ((S i)+ j) by intuition; intuition.
Qed.

Lemma subst_travers : 
 forall  M N P n, (M [← N]) [n ← P] = (M [n+1 ← P])[← N[n← P]].
intros.
rewrite plus_comm. change n with (O+n). apply substP4.
Qed.


Lemma eta_related  : forall s t n m  k l , k+l <= m -> s ↑ n # m = t ↑ k # l -> 
  exists w, s = w ↑ k # l.
induction s; intros.
destruct t; try discriminate.
simpl in H0. 
destruct (le_gt_dec m v); destruct (le_gt_dec l v0).
exists (#(v-k)); simpl.
destruct (le_gt_dec l (v -k)).
rewrite <- le_plus_minus. trivial.
apply le_trans with (k+l); intuition. 
apply le_trans with m; trivial.
apply plus_gt_compat_l with (p:= k) in g.
rewrite <- le_plus_minus in g. elim (lt_irrefl m) .
apply lt_le_trans with (k+l); trivial. apply le_lt_trans with v; intuition.
apply le_trans with (k+l); intuition. 
apply le_trans with m; trivial.
exists (#(v-k)); simpl.
destruct (le_gt_dec l (v -k)).
rewrite <- le_plus_minus. trivial. apply le_trans with (k+l); intuition. 
apply le_trans with m; trivial.
apply plus_gt_compat_l with (p:= k) in g0. rewrite <- le_plus_minus in g0. elim (lt_irrefl m) .
apply lt_le_trans with (k+l); trivial. apply le_lt_trans with v; intuition.
apply le_trans with (k+l); intuition.  apply le_trans with m; trivial.
rewrite H0.
exists #v0; simpl.
destruct (le_gt_dec l v0).
trivial. elim (lt_irrefl v0). apply lt_le_trans with l; trivial.
exists #v0; simpl.
destruct (le_gt_dec l v0).  elim (lt_irrefl v0). apply lt_le_trans with l; trivial.
trivial.
unfold lift_rec in H0; destruct (le_gt_dec m v); discriminate. 
unfold lift_rec in H0; destruct (le_gt_dec m v); discriminate. 
unfold lift_rec in H0; destruct (le_gt_dec m v); discriminate. 
unfold lift_rec in H0; destruct (le_gt_dec m v); discriminate. 
exists !s; simpl; trivial.
destruct t; try discriminate.
unfold lift_rec in H0; destruct (le_gt_dec l v); discriminate.
simpl in H0; injection H0; intros; subst; clear H0.
destruct (IHs3 t3 n m k l H H1); clear IHs3.
destruct (IHs1 t1 n m k l H H3); clear IHs1.
destruct (IHs2 t2 n (S m) k (S l) ); clear IHs2.
replace (k+ S l) with (S (k+l)) by intuition. apply le_n_S. trivial. trivial.
exists (x0·(x1) x); simpl; subst; trivial. 
destruct t; try discriminate.
unfold lift_rec in H0; destruct (le_gt_dec l v); discriminate.
simpl in H0; injection H0; intros; subst; clear H0.
destruct (IHs2 t2 n (S m) k (S l)); intuition; clear IHs2.
replace (k+ S l) with (S (k+l)) by intuition. apply le_n_S. trivial.
destruct (IHs1 t1 n m k l H H2); clear IHs1.
exists (Π(x0), x); simpl; subst; trivial. 
destruct t; try discriminate.
unfold lift_rec in H0; destruct (le_gt_dec l v); discriminate.
simpl in H0; injection H0; intros; subst; clear H0.
destruct (IHs2 t2 n (S m) k (S l));intuition;  clear IHs2.
replace (k+ S l) with (S (k+l)) by intuition. apply le_n_S. trivial.
destruct (IHs1 t1 n m k l H H2); clear IHs1.
exists (λ[x0],x); simpl; subst; trivial.
Qed. 

