Require Import base.
Require Import ut_term.
Require Import ut_red.
Require Import ut_env.
Require Import ut_typ.
Require Import List.

Open Scope UT_scope.

(* How to deal with Beta reduction in the context *)
Inductive Beta_env : Env -> Env -> Prop :=
 | Beta_env_hd : forall A B e, A → B -> Beta_env (A::e) (B::e)
 | Beta_env_cons : forall A e f, Beta_env e f -> Beta_env (A::e) (A::f)
.

Inductive Betas_env : Env -> Env -> Prop :=
 | Betas_env_refl : forall e, Betas_env e e
 | Betas_env_Beta : forall e f, Beta_env e f -> Betas_env e f
 | Betas_env_trans : forall e f g, Betas_env e f -> Betas_env f g -> Betas_env e g.

Inductive Betac_env : Env -> Env -> Prop :=
 | Betac_env_Betas : forall e f, Betas_env e f -> Betac_env e f
 | Betac_env_sym :forall e f, Betac_env e f -> Betac_env f e
 | Betac_env_trans : forall e f g, Betac_env e f -> Betac_env f g -> Betac_env e g.


Hint Constructors Beta_env Betas_env Betac_env.

Notation " a →e b " := (Beta_env a b) (at level 20) : UT_scope.
Notation " a →→e b " := (Betas_env a b) (at level 20) : UT_scope.
Notation " a ≡e b " := (Betac_env a b) (at level 20) : UT_scope.

(* simples facts on env / beta *)
Lemma Betas_env_nil_ : forall e g,e →→e g -> (e = nil -> g = nil).
intros e g H.
induction H; intros.
trivial.
subst.
inversion H.
intuition.
Qed.

Lemma Betas_env_nil : forall g, nil →→e g -> g = nil.
intros.
apply (Betas_env_nil_ nil g H); intuition.
Qed.

Lemma Betas_env_nil2_ : forall e g , e →→e g -> (g = nil -> e = nil).
intros e g H; induction H; intros.
trivial.
subst; inversion H.
intuition.
Qed.

Lemma Betas_env_nil2 : forall e, e →→e nil -> e = nil.
intros ; apply (Betas_env_nil2_ e nil) ; trivial.
Qed.

Lemma Betas_env_inv_ : forall E F, E →→e F -> 
 (forall a b g f,E = a::g -> F = b::f  -> a→→b /\ g→→e f).
intros E F H; induction H; intros.
rewrite H0 in H.
injection H; intros; subst.
split; constructor.
inversion H; subst.
injection H3; injection H4; intros; subst.
rewrite H5; split; constructor; trivial.
injection H3; injection H4; intros; subst.
rewrite H6; split; constructor; trivial.
destruct f.
rewrite (Betas_env_nil g H0) in *; discriminate.
subst.
destruct (IHBetas_env2 t b f f0); intuition; 
destruct (IHBetas_env1 a t g0 f); intuition.
apply Betas_trans with (y:=t); trivial.
apply Betas_env_trans with  (f:=f); trivial.
Qed.

Lemma Betas_env_inv :forall a b g f, (a::g)  →→e  (b::f) -> a →→ b /\ g→→e f.
intros.
apply Betas_env_inv_ with (a::g) (b::f) ; trivial.
Qed.

Lemma Betas_env_hd : forall A C f, A→→C -> (A::f) →→e (C::f).
intros.
induction H.
constructor.
constructor; apply Beta_env_hd; trivial.
apply Betas_env_trans with (f:=y::f); intuition.
Qed.

Lemma Betas_env_cons : forall A g f , g →→e f -> (A::g) →→e (A::f).
intros.
induction H.
constructor.
constructor; apply Beta_env_cons; trivial.
apply Betas_env_trans with (f:=A::f); trivial.
Qed.

Lemma Betas_env_comp : forall A B g f, A→→B -> g→→e f -> (A::g) →→e (B::f).
intros.
apply Betas_env_trans with (f:= A::f).
apply Betas_env_cons; trivial.
apply Betas_env_hd; trivial.
Qed.

(* Important one 
we have to proof subject reduction and context reduction at the same time
*)

Lemma Beta_env_sound :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ' ⊣ -> Γ →e Γ' -> Γ' ⊢ x : A  .
induction 1; intros.
(**)
constructor; trivial.
(**)
generalize dependent A. revert H1 H v. 
induction H2; intros. destruct H2 as (AA&  ? & ?). inversion H3; subst; clear H3.
inversion H1; subst; clear H1. inversion H0; subst; clear H0.
apply Cnv with ( B  ↑ 1 ) s0. eauto.
constructor. eauto.  exists B; intuition.
change !s0 with !s0 ↑ 1. apply thinning with s ;trivial.
constructor. eauto. exists AA; intuition.
destruct H0 as (AA&  ? & ?).  inversion H3; subst; clear H3.
constructor. trivial. exists A; intuition.
change (S(S n)) with (1+S n). change #(S n) with (#n  ↑ 1).
rewrite <- lift_lift.
inversion H1; subst; clear H1.
apply thinning with s; trivial. apply IHBeta_env. eauto. inversion H; subst; clear H. eauto. 
exists AA; intuition.

econstructor; eauto.
econstructor. apply H. eauto. eauto. eauto.
econstructor. eauto. eauto.

eauto.
Qed.



Lemma Beta_env_sound_up :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ' ⊣ -> Γ' →e Γ -> Γ' ⊢ x : A  .
induction 1; intros.
(**)
constructor; trivial.
(**)
generalize dependent A. revert H1 H v. 
induction H2; intros. destruct H2 as (AA&  ? & ?). inversion H3; subst; clear H3.
inversion H1; subst; clear H1. inversion H0; subst; clear H0.
apply Cnv with ( A  ↑ 1 ) s0. eauto.
constructor. eauto.  exists A; intuition.
change !s0 with !s0 ↑ 1. apply thinning with s ;trivial.
constructor. eauto. exists AA; intuition.
destruct H0 as (AA&  ? & ?).  inversion H3; subst; clear H3.
constructor. trivial. exists A; intuition.
change (S(S n)) with (1+S n). change #(S n) with (#n  ↑ 1).
rewrite <- lift_lift.
inversion H1; subst; clear H1.
apply thinning with s; trivial. apply IHBeta_env. eauto. inversion H; subst; clear H. eauto. 
exists AA; intuition.

econstructor; eauto.
econstructor. apply H. eauto. eauto. eauto.
econstructor. eauto. eauto.

eauto.
Qed.

Lemma Beta_sound : forall Γ x A, Γ ⊢ x : A-> forall y, x→ y ->  Γ ⊢y : A.
induction 1; intros.
(*1*)
inversion H1.
(*2*)
inversion H1.
(*3*)
inversion H2; subst; clear H2. econstructor; eauto. econstructor. apply H. intuition.
eapply Beta_env_sound. apply H1. eauto. intuition.
(*4*)
inversion H3; subst; clear H3. eauto.
apply Cnv with ( Π (B0), B) s3. eauto.
econstructor. apply H. eauto.
eapply Beta_env_sound. apply H1. econstructor. eauto. intuition.
eapply Beta_env_sound. apply H2. econstructor. eauto. intuition.
econstructor; eauto.
(*5*)
inversion H1; subst; clear H1.
assert (exists u, Γ ⊢ Π(A),B : !u). apply TypeCorrect in H.
destruct H. destruct H; discriminate. trivial.
destruct H1. apply Geniii in H1 as (s & t & u & h); decompose [and] h; clear h.
apply Geniv in H. destruct H as (u1 & u2 & u3 & D & ? & ? & ? & ? & ?).
apply Betac_Pi_inv in H as (? & ?).
eapply Cnv with (s := t). eapply Betac_subst2. apply Betac_sym. apply H9.
eapply substitution. apply H7. eapply Cnv. apply H. apply H0. apply H6. constructor.
eauto. change !t with (!t  [ ← b] ). eapply substitution. apply H5. apply H0. constructor.
eauto.

econstructor.  eauto. trivial.
assert (exists u, Γ ⊢ Π(A),B : !u). apply TypeCorrect in H.
destruct H. destruct H; discriminate. trivial.
destruct H1. apply Geniii in H1 as (s & t & u & h); decompose [and] h; clear h.
apply Cnv with (B [← y0]) t. apply Betac_subst. auto.
econstructor. apply H. eauto.
change !t with (!t  [ ← b] ). eapply substitution. apply H6. apply H0. constructor.
eauto.
(**)
eauto.
Qed.


Lemma SubjectRed : forall Γ x T, Γ ⊢ x : T-> 
  forall y , x →→ y -> Γ ⊢ y : T.
intros. induction H0.
trivial. eapply Beta_sound.
apply H. trivial. intuition.
Qed.

(* Type reduction *)
Lemma Betas_typ_sound : forall Γ x T, Γ ⊢ x : T -> 
 forall T', T →→ T' -> Γ ⊢ x : T'.
intros.
assert ((exists s, T = !s )\/ exists s,  Γ ⊢ T : !s ) by ( apply TypeCorrect in H; intuition).
destruct H1 as [[s Hs]|[s Hs]].
rewrite Hs in *. apply Betas_S in H0. rewrite H0; trivial.
apply Cnv with T s; intuition.
apply SubjectRed with T; intuition.
Qed.

Lemma Beta_env_sound2 : forall Γ Γ', Γ ⊣ -> Γ →e Γ' -> Γ' ⊣.
intros. induction H0; intuition.
apply wf_inv in H. destruct H as (s & ? ).
econstructor. eapply Beta_sound. apply H. trivial.
inversion H; subst; clear H.
econstructor. eapply Beta_env_sound. apply H2.  apply IHBeta_env.
eauto. trivial.
Qed.

Lemma Betas_env_sound2 : forall Γ Γ', Γ ⊣ -> Γ →→e Γ' -> Γ' ⊣.
intros.
induction H0; intuition.
eapply Beta_env_sound2. apply H. trivial.
Qed.

Lemma Betas_env_sound :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ →→e Γ' -> Γ' ⊢ x : A  .
intros.
induction H0; intuition.
eapply Beta_env_sound. apply H. apply Beta_env_sound2 with e. apply wf_typ in H. trivial.
trivial. trivial.
Qed.

Lemma Betas_env_sound_up :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ' ⊣ -> Γ' →→e Γ -> Γ' ⊢ x : A  .
intros. revert x A H H0 .
induction H1; intros.
trivial.
eapply Beta_env_sound_up. apply H0. trivial. trivial.
apply IHBetas_env1. apply IHBetas_env2. trivial. apply Betas_env_sound2 with e; trivial.
trivial.
Qed.




(* compilation of all this *)
Lemma Subject_Reduction :  forall Γ x A, Γ ⊢ x : A ->
  forall Γ' y B, x→→ y -> Γ →→e Γ' -> A→→ B -> Γ' ⊢ y : B.
intros.
apply Betas_typ_sound with A; intuition.
apply Betas_env_sound with Γ; intuition.
apply SubjectRed with x; intuition.
Qed.


Inductive Tv : Term -> Prop :=
 | Tv_intro : forall x, Tv #x
 | Tv_la: forall b A, Tv b -> Tv (λ[A],b)
 | Tv_app :forall b c, Tv b -> Tv (b·c)
.

Inductive Ts : Term -> Prop :=
 | Ts_intro : forall s, Ts !s
 | Ts_pi : forall A B, Ts (Π(A),B)
 | Ts_la : forall b A, Ts b -> Ts (λ[A],b)
 | Ts_app : forall b c, Ts b -> Ts (b·c)
.

Hint Constructors Tv Ts.

Require Import Arith.

Lemma Tv_lift : forall M n m, Tv M <-> Tv (M ↑ n # m).
induction M; split; simpl in *; intros; intuition.
destruct le_gt_dec; intuition.
inversion  H; subst; clear H. constructor. eapply IHM1. trivial.
inversion  H; subst; clear H. constructor. eapply IHM1. apply H1.
inversion H. inversion H.
inversion  H; subst; clear H. constructor. eapply IHM2. trivial.
inversion  H; subst; clear H. constructor. eapply IHM2. apply H1.
Qed.

Lemma Ts_lift : forall M n m, Ts M <-> Ts (M ↑ n # m).
induction M; split; simpl in *; intros; intuition.
inversion H. destruct le_gt_dec; inversion H.
inversion  H; subst; clear H. constructor. eapply IHM1. trivial.
inversion  H; subst; clear H. constructor. eapply IHM1. apply H1.
inversion  H; subst; clear H. constructor. eapply IHM2. trivial.
inversion  H; subst; clear H. constructor. eapply IHM2. apply H1.
Qed.

Lemma Tx_ok : forall t, Tv t \/ Ts t.
induction t; intros.
left; trivial.
right; trivial.
destruct IHt1.
left; constructor; trivial.  right; constructor; trivial.
right; trivial.
destruct IHt2.
left; constructor; trivial.  right; constructor; trivial.
Qed.


Lemma fun_item_lift : forall A A' v Γ , A ↓ v ⊂ Γ -> A' ↓ v ⊂ Γ -> A = A'.
intros.
destruct H as (x & ?& ?). destruct H0 as (x' & ?& ?).
subst. replace x' with x. trivial. eapply fun_item. apply H1. apply H2.
Qed.

Lemma TU1 : forall Γ M A , Γ ⊢ M : A -> forall A',  Γ ⊢ M : A' -> Tv M ->  A ≡ A'.
induction 1; intros.
inversion H2.
(**)
apply Genii in H1 as (B & ? &?).
replace A with B. eauto. eapply fun_item_lift. apply H3. trivial.
(**)
inversion H3; subst; clear H3.
(**)
inversion H4; subst; clear H4.
apply Geniv in H3 as (t1 & t2 & t3 & D & h); decompose [and] h;clear h.
eapply Betac_trans. eapply Betac_Pi. apply Betac_refl. eapply IHtyp3. apply H7. trivial.
intuition.
(**)
inversion H2; subst; clear H2.
apply Genv in H1 as (C & D & ? & ?& ?).
destruct (Betac_Pi_inv A B D C). apply IHtyp1. trivial. trivial.
eapply Betac_trans. eapply Betac_subst2. apply H6. intuition.
(**)
eauto.
Qed.


Lemma foo0 : forall l A B  t, B →→ Tels l t -> Π (A), B →→ Tels (A::l) t.
induction l; intros; simpl in *. 
eauto.
eauto.
Qed.

Lemma Beta_Tels_inv : forall l t T, Tels l t → T -> (exists l', T = Tels l' t /\ l →e l') \/ 
  (exists t', T = Tels l t' /\ t → t').
induction l; intros; simpl in *.
right. exists T; intuition.
inversion H; subst; clear H.
destruct (IHl t y H3). destruct H as (l' & -> & ?).
left. exists (a::l'); intuition.
destruct H as (t' & -> & ?).
right. exists t'; intuition.

left. exists (B::l); intuition.
Qed.


Lemma Betas_Tels_inv : forall l t T, Tels l t →→ T -> exists l', exists t', T = Tels l' t' /\ l →→e l' /\ t →→ t'.
intros.
remember (Tels l t) as TT. revert l t HeqTT.
induction H; intros; subst.
exists l; exists t; intuition.
apply Beta_Tels_inv in H.
destruct H. destruct H as (l' & -> & ?).
exists l'; exists t; intuition.
destruct H as (t' & -> & ?).
exists l; exists t'; intuition.
destruct (IHBetas1 l t) as (l' & t' & -> & ?& ?); trivial.
destruct (IHBetas2 l' t') as (l'' & t'' & -> & ?& ?); trivial.
exists l''; exists t''; eauto.
Qed.


Lemma Betas_Tels : forall l l' t t', l →→e l' -> t →→ t' -> Tels l t →→ Tels l' t'.
assert (forall l l', l →→e l' -> forall t ,  Tels l t →→ Tels l' t).
induction 1; intros; eauto.
induction H; simpl; eauto.
assert (forall t t', t →→ t' -> forall l ,  Tels l t →→ Tels l t').
induction 1; intros; eauto.
induction l; simpl; eauto.
intros.
eauto.
Qed.

Lemma Betas_not_Pi_to_Sort : forall A B s,  ~ (Π (A), B →→ !s).
intros; intro. eapply Betac_not_Pi_sort. constructor. apply H.
Qed.


Fixpoint env_subst E n N {struct E } := match E with
 | nil => nil 
 | A::F => (A[n ← N]) :: (env_subst F (S n) N)
end.

Notation "E  [[ n ← N ]]" := (env_subst E n N) (at level 80).

Lemma Tels_subst : forall l t n N, (Tels l t) [n ← N] = Tels (l [[n ← N]]) (t [(length l + n) ← N]).
induction l; simpl; intros.
trivial.
rewrite IHl.
rewrite plus_n_Sm. trivial.
Qed.

Lemma Tels'_subst : forall l t n N, (Tels' l t) [n ← N] = Tels' (l [[n ← N]]) (t [(length l + n) ← N]).
induction l; simpl; intros.
trivial.
rewrite IHl.
rewrite plus_n_Sm. trivial.
Qed.



Lemma TU2 : forall Γ M A,  Γ ⊢ M : A -> forall A', Γ ⊢ M : A' -> Ts M -> 
  exists s, exists s', exists l,  A →→ (Tels l !s) /\ A' →→ (Tels l !s').
induction 1; intros.
apply Geni in H1 as (s' & ?& ?).
apply Betac_sym in H1. apply conv_to_sort in H1.
exists t; exists s'; exists nil; simpl; intuition.
(**)
inversion H2.
(**)
clear IHtyp1 IHtyp2.
apply Geniii in H2 as (s1 & s2 & s3 & h); decompose [and] h; clear h.
apply Betac_sym in H2. apply conv_to_sort in H2.
exists u; exists s3; exists nil; simpl; intuition.
(**)
inversion H4; subst; clear H4.
apply Geniv in H3 as (t1 & t2 & t3 & D & h); decompose [and] h; clear h.
destruct (IHtyp3 D H7 H6) as (u & v & l & ?& ?).
apply Betac_confl in H3 as (Z & ? &?).
apply Betas_Pi_inv in H11 as (K & L & -> & ? & ?).
destruct (Betas_diamond  D L (Tels l !v)) as ( Z & ?& ?); trivial.
apply Betas_Tels_inv in H14 as (l'' & ? & -> & ? &?).
apply Betas_S in H15; subst.
exists u; exists v; exists (K::l''); split.
apply Betas_trans with (Tels (A::l) !u). simpl. eauto. simpl.
apply Betas_Pi. trivial. apply Betas_Tels; trivial.
simpl. eauto.
(**)
inversion H2; subst; clear H2. clear IHtyp2.
apply Genv in H1 as (C & D & ? & ?&  ?). destruct (IHtyp1 ( Π (D),C) ) as (u & v & l & ? & ?); trivial.
clear IHtyp1. destruct l; simpl in *. apply Betas_not_Pi_to_Sort in H5; elim H5.
apply Betas_Pi_inv in H5 as (B0 & A0 & ? & ?& ?). injection H5; intros;subst; clear H5.
apply Betas_Pi_inv in H6 as (D0 & C0 & ? & ?& ?). injection H5; intros;subst; clear H5.
apply Betac_confl in H1 as (Z & ?& ?).
destruct (Betas_diamond (C[←b]) Z ((Tels l !v)[←b]) H5) as (ZZ & ?& ?).
apply Betas_subst2; trivial.  rewrite Tels_subst in H11. simpl in H11.
apply Betas_Tels_inv in H11. destruct H11 as (l' & x & -> & ?& ?). apply Betas_S in H12. subst.
exists u ; exists v; exists l'; split. eapply Betas_trans. apply Betas_subst2.  apply H8.
rewrite Tels_subst. eapply Betas_Tels. trivial. simpl; trivial.
eauto.
(**)
destruct (IHtyp1 A' H2 H3) as (u & v & l & ? & ?).
apply Betac_confl in H as (Z & ? & ?). destruct (Betas_diamond A (Tels l !u) Z H4 H) as (ZZ & ?& ?).
apply Betas_Tels_inv in H7 as (l' & ? & -> & ? & ?). apply Betas_S in H9; subst.
exists u; exists v; exists l'; split.
apply Betas_trans with Z; trivial. apply Betas_trans with (Tels l !v); trivial.
apply Betas_Tels; trivial.
Qed.


(* jutting suite *)

Inductive Sigma : Env -> Term -> Sorts -> Prop :=
 | Sigma_sorts : forall Γ s s', Ax s s' -> Sigma Γ !s s'
 | Sigma_pi : forall Γ A B s t u, Rel s t u ->
   ( Tv A -> Γ ⊢ A : !s) -> (Ts A -> Sigma Γ A s) ->
   ( Tv B -> A::Γ ⊢ B : !t) -> (Ts B -> Sigma (A::Γ) B t) -> Sigma Γ (Π(A),B) u
 | Sigma_la : forall Γ A b s, Sigma (A::Γ) b s -> Sigma Γ (λ[A],b) s
 | Sigma_app : forall Γ M N s, Sigma Γ M s -> Sigma Γ (M· N) s.

Hint Constructors Sigma.

Lemma Tels_tool : forall Γ Γ' s t, Tels Γ !s →→ Tels Γ' !t -> s = t /\ length Γ = length Γ'.
induction Γ; destruct Γ'; simpl in *; intros.
apply Betas_S in H. injection H; intros; subst; intuition.
apply Betas_S in H; discriminate.
apply Betas_not_Pi_to_Sort in H; elim H.
apply Betas_Pi_inv in H as ( C & D & ? & ? & ?). injection H ; intros; subst; clear H.
destruct (IHΓ Γ' s t0 H1). subst; intuition. 
Qed.

Lemma L52 : forall a Z Δ A n Γ Γ' d, Γ ⊢ a : Z -> Δ ⊢ A : !d -> Ts a -> ins_in_env Δ A n Γ Γ' -> forall s , Sigma Γ a s <-> Sigma Γ' (a↑1#n) s.
induction a; intros; simpl in *.
(**)
inversion H1.
(**)
split; intros. inversion H3; subst; clear H3. intuition. inversion H3; subst; clear H3. intuition.
(**)
inversion H1; subst; clear H1. apply Genv in H. destruct H as ( C & D & ? & ?& ?). 
destruct (IHa1 (Π(D),C) Δ A n Γ Γ' d H1 H0 H4 H2 s).
split; intros. inversion H7; subst; clear H7. intuition.  inversion H7; subst; clear H7. intuition.
(**)
apply Geniii in H. destruct H as (u & v & w & ? & ? & ? & ?).
split; intros. inversion H6; subst; clear H6. econstructor. apply H9. 
intros. apply Tv_lift in H6. change !s0 with (!s0 ↑ 1 # n). eapply weakening. apply (H10 H6). apply H2. apply H0. 
intros. apply Ts_lift in H6. eapply IHa1. apply H4. apply H0. trivial. apply H2. intuition.
intros. apply Tv_lift in H6. change !t with (!t ↑ 1 # (S n)). eapply weakening. apply (H13 H6). constructor; apply H2. apply H0.
intros. apply Ts_lift in H6. eapply IHa2. apply H5. apply H0. trivial. constructor; apply H2. intuition.

inversion H6; subst; clear H6. econstructor. apply H9.
intros. replace u with s0 in *. trivial. apply conv_sort. eapply TU1. apply H10. apply Tv_lift. trivial. change !u with (!u ↑ 1 # n).
eapply weakening. apply H4. apply H2. apply H0. apply Tv_lift; trivial.
intros. eapply IHa1. apply H4. apply H0. trivial. apply H2. apply H11. apply Ts_lift; trivial.
intros. replace v with t in *. trivial. apply conv_sort. eapply TU1. apply H13. apply Tv_lift. trivial. change !v with (!v ↑ 1 # (S n)).
eapply weakening. apply H5. constructor; apply H2. apply H0. apply Tv_lift; trivial.
intros. eapply IHa2. apply H5. apply H0. trivial. constructor; apply H2. apply H15. apply Ts_lift; trivial.
(**)
inversion H1; subst; clear H1. apply Geniv in H. destruct H as ( u & v & w &  D & ? & ?& ? & ? & ?). 
destruct (IHa2 D Δ A (S n) (a1::Γ) (a1↑1#n :: Γ') d H5 H0 H4) with (s := s). constructor; trivial.
split; intros. inversion H9; subst; clear H9. intuition.  inversion H9; subst; clear H9. intuition.
Qed.

Lemma L53 : forall Γ a A, Γ ⊢ a : A -> forall Δ s, Ts a -> A →→ Tels Δ !s -> Sigma Γ a s.
induction 1; intros.
(**)
apply Betas_S in H2. destruct Δ; simpl in H2; try discriminate. injection H2; intros; subst; clear H2.
intuition.
(**)
inversion H1.
(**)
apply Betas_S in H3. destruct Δ; simpl in H3; try discriminate. injection H3; intros; subst; clear H3.
apply Sigma_pi with s t; intros; trivial.
apply IHtyp1 with nil; simpl; intuition. apply IHtyp2 with nil; simpl; intuition.
(**)
inversion H3; subst; clear H3.
apply Betas_Pi_inv in H4 as (A' & ? & ? & ?& ?).
destruct Δ; simpl in H3. discriminate. injection H3; intros; subst; clear H3.
constructor. apply IHtyp3 with Δ; trivial.
(**)
inversion H1; subst; clear H1.
destruct (TU2 Γ a (Π(A),B) H (Π(A),B) H H4) as (s' & _ & Θ & ? & _ ).
apply Betas_Pi_inv in H1 as (A' & ? & ? & ?& ?).
destruct Θ; simpl in H1. discriminate. injection H1; intros; subst; clear H1.
replace s with s' in *.
constructor. apply IHtyp1 with (A::Θ); simpl; intuition.
destruct (Betas_diamond (B[← b]) (Tels Δ !s) (Tels Θ !s')[← b] H2) as (Z & ?& ?).
intuition. apply Betas_Tels_inv in H1 as (Δ' & ? & -> & ? & ?).
rewrite Tels_subst in H6. apply Betas_S in H7; subst. simpl in H6.
apply Tels_tool in H6; intuition.
(**)
destruct (Betac_confl A B H) as ( Z  & ?& ?). destruct (Betas_diamond B Z (Tels Δ !s0) H5 H3) as (ZZ & ?& ?).
apply Betas_Tels_inv in H7 as (Θ & ? & -> & ? & ?). apply Betas_S in H8; subst.
eauto.
Qed.

Lemma L54 : forall Γ A s, Γ ⊢ A : !s -> Ts A -> Sigma Γ A s.
intros. apply L53 with !s nil; simpl; trivial.
Qed.


Lemma env_subst_tool : forall Γ n M, length (Γ[[n ← M]]) = length Γ.
induction Γ; simpl in *; intros; intuition.
Qed.

Lemma foo1 : forall Γ Δ Δ', Δ →→e Δ' -> (Γ++Δ) →→e (Γ++Δ').
induction Γ; simpl in *; intros.
trivial.
apply Betas_env_cons. intuition.
Qed.


Fixpoint env_subst_down E N n {struct E } := match E with
 | nil => nil 
 | A::F => (A[length F + n ← N]) :: (env_subst_down F N n)
end.

Lemma test0 :forall Γ Γ' t n,  env_subst_down (Γ++Γ') t n = (env_subst_down Γ t (length Γ' + n))++(env_subst_down Γ' t n).
induction Γ; simpl in *; intros. trivial.
replace (length (Γ++Γ')+n) with (length Γ + (length Γ' + n)).
rewrite IHΓ. trivial. rewrite app_length. intuition.
Qed.

Lemma test1 : forall Θ t n, rev (Θ [[n ← t]]) = env_subst_down (rev Θ) t n.
induction Θ; simpl in *; intros. trivial.
rewrite test0. rewrite IHΘ. simpl. trivial.
Qed.

Lemma test2 : forall Θ Γ t A, sub_in_env Γ t A (length Θ) (Θ ++ A :: Γ) (env_subst_down Θ t 0 ++ Γ).
induction Θ; simpl in * ;intros. constructor.
replace (length Θ+0) with (length Θ) by intuition. apply sub_S. trivial.
Qed.

Lemma foo2 : forall Θ Γ t A ,  sub_in_env Γ t A (length Θ) (rev (A :: Θ) ++ Γ) (rev (Θ [[0 ←t]]) ++ Γ).
intros. rewrite test1. simpl. rewrite <- app_assoc;  simpl. rewrite <- rev_length. apply test2.
Qed.



Lemma foo3 : forall Γ Γ', Γ →→e Γ' -> length Γ = length Γ'.
induction 1; intuition. induction H; simpl; intuition.
Qed.


Lemma L55 : forall Γ a A, Γ ⊢ a : A -> forall Δ s s1, Ts a -> A →→ Tels Δ !s1 -> Sigma Γ a s -> 
  exists Δ', exists a1, length Δ = length Δ' /\ a →→ Tels' Δ' a1 /\ rev Δ'++Γ ⊢ a1 : !s.
induction 1; intros.
(**)
exists nil; exists !s; simpl; intuition.
destruct Δ; simpl in H2. trivial. apply Betas_S in H2; discriminate.
inversion H3; subst; clear H3. intuition.
(**)
inversion H1.
(**)
replace Δ with (@nil Term) in *. simpl in *.
inversion H4; subst; clear H4.
destruct (Tx_ok A).
  destruct (Tx_ok B).
    exists nil; exists (Π(A),B); simpl; intuition. econstructor. apply H7. trivial. trivial.
    
    destruct (IHtyp2 nil t0 t H5 ) as (Δ' & b & ? & ? & ?). simpl; trivial. intuition.
    replace Δ' with (@nil Term) in *; simpl in *. exists nil; exists (Π(A), b); simpl; intuition.
    econstructor. apply H7. trivial. trivial. destruct Δ'. trivial. discriminate.
  destruct (IHtyp1 nil s2 s) as (Δ' & a & ? & ?& ?). trivial. simpl; trivial. intuition.
  replace Δ' with (@nil Term) in *; simpl in *. destruct (Tx_ok B).
    exists nil; exists (Π(a),B); simpl; intuition. econstructor. apply H7. trivial.
    eapply Betas_env_sound. apply H9. apply Betas_env_comp; intuition.

    destruct (IHtyp2 nil t0 t H12) as (Δ'' & b & ? & ? & ?). simpl; trivial. intuition.
    replace Δ'' with (@nil Term) in *; simpl in *. exists nil; exists (Π(a),b); simpl; intuition.
    econstructor. apply H7. trivial. eapply Betas_env_sound. apply H16. apply Betas_env_comp; intuition.
    destruct Δ''. trivial. discriminate.     destruct Δ'. trivial. discriminate.
    apply Betas_S in H3. destruct Δ. trivial. discriminate.
(**)
inversion H3; subst; clear H3. inversion H5; subst; clear H5. apply Betas_Pi_inv in H4 as (A' & B' & ? & ?& ?).
destruct Δ. discriminate. injection H3; intros; subst; clear H3.
destruct (IHtyp3 Δ s s0 H7 H5 H10) as (Δ' & b' & ? & ? & ?).
exists (A::Δ'); exists b'; simpl; intuition. rewrite <- app_assoc. simpl. trivial.
(**)
inversion H3; subst; clear H3. inversion H1; subst; clear H1.
destruct (TU2 Γ a (Π(A),B) H (Π(A),B) H H4) as ( sa & _ & Θ & ? & _ ).
apply Betas_Pi_inv in H1 as (A' & B' & ? & ?& ?). destruct Θ. discriminate.
injection H1; intros; subst; clear H1. replace sa with s1 in *.
destruct (IHtyp1 (A::Θ) s s1 H4) as ( Θ' & m & ? & ? &?). simpl; intuition. trivial.
destruct Θ' as [ | A'' Θ']. discriminate. simpl in H1. injection H1; intros; subst; clear H1.
exists (Θ'[[ 0 ← b]]); exists (m[ (length Θ' )← b]); repeat split.
rewrite env_subst_tool.   destruct (Betas_diamond (B[← b]) (Tels Δ !s1) (Tels Θ !s1)[← b] H2 ) as (Z & ?& ?).
intuition. apply Betas_Tels_inv in H1 as (? & ?& -> & ?& ?).
rewrite Tels_subst in H10. apply Betas_S in H11; subst.
apply Tels_tool in H10 as ( _ & ?). replace (length Δ) with (length x).
rewrite env_subst_tool in H10. intuition. apply foo3 in H1. intuition. apply Betas_trans with ((λ[A''],Tels' Θ' m)· b).
intuition. replace (length Θ') with (length Θ' + 0). rewrite <- Tels'_subst. constructor.
constructor. intuition. assert (A≡ A'' /\ exists s, Γ ⊢ A'' : !s).
assert ( Γ ⊢ λ[A''],Tels' Θ' m : Π(A),B). eapply SubjectRed. apply H. trivial.
  apply Geniv in H1. destruct H1 as ( ss & ? & ? & ? & ? & _ & ? & _ ). split.
  apply Betac_Pi_inv in H1; intuition. exists ss; trivial.
change !s with (!s[ (length Θ') ← b]).   destruct H1 as (? & ss & ?). eapply substitution. apply H7.  eapply Cnv. apply H1. apply H0.
apply H10. apply foo2. apply wf_typ in H7; trivial.  destruct (Betas_diamond (B[← b]) (Tels Δ !s1) (Tels Θ !sa)[← b] H2) as (Z & ?& ?).
intuition. apply Betas_Tels_inv in H1 as (Δ' & ? & -> & ? & ?).
rewrite Tels_subst in H6. apply Betas_S in H7; subst. simpl in H6.
apply Tels_tool in H6; intuition.
(**)
destruct (Betac_confl A B H) as (Z & ? & ?). destruct (Betas_diamond B (Tels Δ !s1) Z H3 H6) as (ZZ & ? &?).
apply Betas_Tels_inv in H7 as (? & ? & -> & ? & ?). apply Betas_S in H9; subst.
destruct (IHtyp1 x s0 s1 H2) as ( Δ' & a1 & ? & ? & ?). eauto. trivial.
exists Δ'; exists a1; intuition. rewrite <- H9. apply foo3 in H7. trivial.
Qed.

Lemma L56 : forall Γ A s, Γ ⊢ A : !s -> forall s', Ts A -> Sigma Γ A s' -> exists A', A →→ A' /\ Γ ⊢ A' : !s'.
intros.
destruct (L55 Γ A !s H nil s' s H0) as ( ? & a & ? & ? & ?). simpl; trivial. trivial.
replace x with (@nil Term) in *; simpl in *.
exists a; intuition. destruct x. trivial. discriminate.
Qed.

Lemma L1 : forall Δ Z n Γ0 Γ, ins_in_env Δ Z n Γ0 Γ -> forall v a, n <= v -> a ↓ S v ∈ Γ ->
  a ↓ v ∈ Γ0.
induction 1; intros. inversion H0; subst; clear H0. trivial.
inversion H1; subst; clear H1. destruct v. contradict H0; intuition.
apply le_S_n in H0. constructor. intuition.
Qed.


Lemma L2 : forall Δ Z n Γ0 Γ, ins_in_env Δ Z n Γ0 Γ -> forall v a, n > v ->  a ↓ v ∈ Γ ->
 exists b, b ↓ v ∈ Γ0 /\ b↑1# (n-S v) = a.
induction 1; intros. contradict H; intuition. inversion H1; subst; clear H1.
exists d; intuition. replace (S n - 1) with n by intuition. trivial.
apply lt_S_n in H0. destruct (IHins_in_env n0 a) as ( b & ? & ?). intuition. trivial.
exists b; split. intuition. replace (S n - S (S n0)) with (n - S n0) by intuition.
trivial.
Qed.


Lemma L3 : forall Δ A n Γ Γ', ins_in_env Δ A n Γ Γ' -> Γ' ⊣ -> exists s, Δ ⊢ A : !s.
induction 1; intros.
inversion H; subst; clear H. exists s; trivial.
inversion H0; subst; clear H0. apply wf_typ in H2. intuition.
Qed.

Lemma L61: (forall Γ' M T, Γ' ⊢ M : T -> forall m n Δ A Γ, M = m ↑ 1 # n -> 
  ins_in_env Δ A n Γ Γ' ->  exists T', T →→ T'↑ 1 # n /\ Γ ⊢ m : T' ) /\
  (forall Γ', Γ' ⊣ -> forall n Δ A Γ, ins_in_env Δ A n Γ Γ' -> Γ ⊣).
apply typ_induc; intros.
(**)
destruct m; simpl in H0; try discriminate. destruct le_gt_dec; discriminate. injection H0; intros; subst; clear H0.
exists !t; split; trivial. constructor. trivial. eauto.
(**)
destruct m; simpl in H0; try discriminate. destruct le_gt_dec; injection H0; intros; subst; clear H0.
destruct i as ( a & ? & ?).   exists (a ↑ (S v0)); split. rewrite liftP3; subst; intuition.
constructor. eauto. exists a; intuition. eapply L1. apply H1. trivial. trivial.

destruct i as (a & ? & ?). destruct (L2 Δ A0 n Γ0 Γ H1 v0 a g H2) as ( b& ? & ?). exists (b ↑ (S v0)); split.
replace n with (S v0 + (n - S v0)) by intuition. rewrite liftP2; intuition. rewrite H4. rewrite H0. intuition.
constructor. eauto. exists b; intuition.
(**)
destruct m; simpl in H1; try discriminate. destruct le_gt_dec; discriminate. injection H1; intros; subst; clear H1.
destruct (H m1 n Δ A0 Γ0) as (A' & ? & ?); trivial. replace A' with !s in *. clear H1; simpl in *. clear H.
destruct (H0 m2 (S n) Δ A0 (m1::Γ0)) as (B' & ? & ?); trivial. constructor. trivial. replace B' with !t in *. clear H; simpl in *. clear H0.
exists !u; intuition. apply cPi with s t; trivial.
apply Betas_S in H. destruct B'; try discriminate. unfold lift_rec in H; destruct le_gt_dec; discriminate. intuition.
apply Betas_S in H1. destruct A'; try discriminate. unfold lift_rec in H1; destruct le_gt_dec; discriminate. intuition.
(**)
destruct m; simpl in H2; try discriminate. destruct le_gt_dec; discriminate. injection H2; intros; subst; clear H2.
destruct (H m1 n Δ A0 Γ0) as (A' & ? & ?); trivial. replace A' with !s1 in *; simpl in *. clear H2. clear H.
destruct (H1 m2 (S n) Δ A0 (m1::Γ0)) as (M & ? & ?); trivial. constructor; trivial. clear H1.
assert (m1↑1#n::Γ ⊢ M ↑ 1 # (S n) : !s2). eapply SubjectRed. apply t0. trivial.
destruct (TypeCorrect (m1::Γ0) m2 M H2) as [ [w ?] | [w ?] ]. 
  subst; simpl in H1. apply Geni in H1. destruct H1 as (ss & ?  & ?). assert (m1::Γ0 ⊢ !w : !s2).
  apply conv_sort in H1. subst. constructor. trivial. apply wf_typ in H2; trivial.
  exists (Π(m1),!w);simpl; intuition. apply cLa with s1 s2 s3; trivial.

  assert (m1 ↑ 1 # n :: Γ ⊢ M ↑ 1 # (S n) : !w). change !w with (!w↑ 1 # (S n)). destruct (L3 Δ A0 n Γ0 Γ H3) as (a & ? ).
  apply wf_typ in t; trivial. eapply weakening. apply H5. constructor. apply H3. apply H6.
  destruct (Tx_ok (M↑ 1 # (S n))). replace w with s2 in *. exists (Π(m1),M); simpl; intuition.
  apply cLa with s1 s2 s3; trivial. apply conv_sort. eapply TU1. apply H1. trivial. trivial. 
  destruct (L56 (m1::Γ0) M w H5 s2) as ( M' & ? & ?). apply Ts_lift in H7; trivial.
  destruct (L3 Δ A0 (S n) (m1::Γ0) (m1↑1#n::Γ)) as (d & ?). constructor; trivial. apply wf_typ in H1; trivial.
  eapply L52. apply H5. apply H8. apply Ts_lift in H7; trivial. constructor. apply H3. apply L54; trivial.
  exists (Π(m1),M'); simpl; split. eauto. apply cLa with s1 s2 s3; trivial. eauto.
  apply Betas_S in H2. destruct A'; try discriminate. unfold lift_rec in H2; destruct le_gt_dec; discriminate. intuition.
(**)
destruct m; simpl in H1; try discriminate. destruct le_gt_dec; discriminate. injection H1; intros; subst; clear H1.
destruct (H m1 n Δ A0 Γ0) as (P & ? & ?); trivial. apply Betas_Pi_inv in H1 as (C & D & ? & ?& ?). 
destruct P; try discriminate. unfold lift_rec in H1; destruct le_gt_dec; discriminate. injection H1; intros; subst; clear H1.
destruct (H0 m2 n Δ A0 Γ0) as (A' & ? & ?); trivial. destruct (Betas_diamond A (P1 ↑ 1 # n) (A' ↑ 1 # n) H4 H1) as ( Z & ?& ?).
apply Betas_lift_inv in H7. destruct H7 as (ZZ & ? & -> ).
apply Betas_lift_inv in H8. destruct H8 as (? & ? & ?). apply inv_lift in H9. subst.
exists (P2 [← m2 ]); split. change n with (0+n). rewrite substP1. intuition. apply cApp with x.
eapply Betas_typ_sound. apply H3. intuition. eapply Betas_typ_sound. apply H6. intuition.
(**)
destruct (H m n Δ A0 Γ0 H1 H2) as (a' & ? & ?).
destruct (Betac_confl A B b) as ( Z & ? &?). destruct (Betas_diamond A (a'↑ 1 # n) Z H3 H5) as (ZZ & ?& ?).
apply Betas_lift_inv in H7. destruct H7 as (aa & ? & -> ). exists aa; split. eauto.
eapply Betas_typ_sound. apply H4. trivial.
(** wf **)
inversion H.
(**)
inversion H0; subst; clear H0.
apply wf_typ in t; trivial.
destruct (H d n0 Δ A0 Δ0) as (? & ? & ?); trivial.
apply Betas_S in H0. destruct x; try discriminate. unfold lift_rec in H0; destruct le_gt_dec; discriminate.
eauto.
Qed.


Lemma Strenghthening: forall Γ' M n T, Γ' ⊢ M ↑ 1 # n: T ↑ 1 # n -> forall Δ A Γ, ins_in_env Δ A n Γ Γ' ->  
 Γ ⊢ M : T.
intros. destruct L61 as ( ? & _ ). destruct (H1 Γ' (M↑ 1 # n) (T↑ 1 # n) H M n Δ A Γ) as (T' & ? & ?); trivial.
apply Betas_lift_inv in H2 as ( ? & ?& ?). apply inv_lift in H4; subst.
apply TypeCorrect in H as [ [ w ?]|[ w ?] ]. destruct T ; try discriminate. unfold lift_rec in H; destruct le_gt_dec; discriminate.
apply Betas_S in H2; subst. trivial. destruct (H1 Γ' (T↑1#n) !w H T n Δ A Γ) as (T'' & ? & ?); trivial. apply Betas_S in H4. 
destruct T'' ; try discriminate. unfold lift_rec in H4; destruct le_gt_dec; discriminate. eauto.
Qed.
  
  



